"""A utility for timing."""

import time

import numpy as np


class Timer(object):

    def __init__(self, task_name="UntitledTask"):
        self.task_name = task_name
        self._duration_list = []
        self.now = None
        self.check_point = None
        self.is_timing = False
        self._count = 0

    def start(self):
        if not self.is_timing:
            self.check_point = time.time()
            self.is_timing = True

    def pause(self):
        if self.is_timing:
            self._duration_list.append(time.time() - self.check_point)
            self.is_timing = False
            self._count += 1

    def stop(self):
        self.pause()
        self.report()

    def report(self):
        print("[Timer] {} total: {:.4f} mean: {:.4f} count: {}".format(
            self.task_name, np.sum(self._duration_list),
            np.mean(self._duration_list), self._count))

    @property
    def duration(self):
        return np.sum(self._duration_list)

    @property
    def count(self):
        return self._count



"""Methods to compute common machine learning metrics"""

import numpy as np


def accuracy(predictions, targets):
    total_num = len(predictions)
    hit_num = int(np.sum(predictions == targets))
    return {"total_num": total_num,
            "hit_num": hit_num,
            "accuracy": 1.0 * hit_num / total_num}


def log_loss(predictions, targets):
    assert len(predictions) == len(targets)
    total_num = len(predictions)

    predictions = np.asarray(predictions)
    targets = np.asarray(targets)

    total_log = np.log(predictions[range(total_num), targets])
    logloss = 1.0 - np.mean(total_log)
    return {"log_loss": logloss}


def precision(predictions, targets, pos_class=1, neg_class=0):
    """precision = TP / (TP + FP)"""
    true_pos, false_pos = 0, 0
    for pred, tar in zip(predictions, targets):
        if pred == pos_class and tar == pos_class:
            true_pos += 1.0
        elif pred == pos_class and tar == neg_class:
            false_pos += 1.0
    precision = true_pos / (true_pos + false_pos)
    return {"precision": precision, "true_positive": true_pos, 
            "false_positive": false_pos}


def recall(predictions, targets, pos_class=1, neg_class=0):
    """recall = TP / (TP + FN)"""
    true_pos, false_neg = 0, 0
    for pred, tar in zip(predictions, targets):
        if pred == pos_class and tar == pos_class:
            true_pos += 1.0
        elif pred == neg_class and tar == pos_class:
            false_neg += 1.0
    recall = true_pos / (true_pos + false_neg)
    return {"recall": recall, "true_positive": true_pos,
            "false_negative": false_neg}


def f1(predictions, targets, pos_class=1, neg_class=0):
    p = precision(predictions, targets, pos_class, neg_class)
    r = recall(predictions, targets, pos_class, neg_class)
    return {"f1": 2 * (p * r) / (p + r), "precision": p, "recall": r}


def explained_variation(predictions, targets):
    """
    Computes fraction of variance that pred_y explains about y.
    Returns 1 - Var[y-pred_y] / Var[y]

    Interpretation:
        EV=0  =>  might as well have predicted zero
        EV=1  =>  perfect prediction
        EV<0  =>  worse than just predicting zero
    """
    assert predictions.shape == targets.shape
    if predictions.ndim == 1:
        diff_var = np.var(targets - predictions)
        target_var = np.var(targets)
    elif predictions.ndim == 2:
        diff_var = np.var(targets - predictions, axis=0)
        target_var = np.var(targets, axis=0)

    non_zero_idx = np.where(target_var != 0)[0]

    ev = np.mean(1.0 - diff_var[non_zero_idx] / target_var[non_zero_idx])
    return {"mean_ev": ev}


def r_square(predictions, targets):
    assert predictions.shape == targets.shape
    ss_residual = np.sum(targets - predictions, axis=0)
    ss_total = np.sum(targets - np.mean(targets, axis=0), axis=0)
    r2 = np.mean(1 - ss_residual / ss_total)
    return {"r_square": r2}


def mean_square_error(predictions, targets):
    assert predictions.shape == targets.shape
    if predictions.ndim == 1:
        mse = np.mean(np.square(predictions - targets))
    elif predictions.ndim == 2:
        mse = np.mean(np.sum(np.square(predictions - targets), axis=1))
    else:
        raise ValueError("predictions supposes to have 1 or 2 dim.")
    return {"mse": mse}


def mean_absolute_error(predictions, targets):
    assert predictions.shape == targets.shape
    if predictions.ndim == 1:
        mae = np.mean(np.abs(predictions - targets))
    elif predictions.ndim == 2:
        mae = np.mean(np.sum(np.abs(predictions - targets), axis=1))
    else:
        raise ValueError("predictions supposes to have 1 or 2 dim.")
    return {"mae": mae}



"""Model class manage the network, loss function and optimizer."""

import pickle


class Model(object):

    def __init__(self, net, loss, optimizer):
        self.net = net
        self.loss = loss
        self.optimizer = optimizer

    def forward(self, inputs):
        return self.net.forward(inputs)

    def backward(self, preds, targets):
        loss = self.loss.loss(preds, targets)
        grad_from_loss = self.loss.grad(preds, targets)
        struct_grad = self.net.backward(grad_from_loss)
        return loss, struct_grad

    def apply_grads(self, grads):
        params = self.net.params
        self.optimizer.step(grads, params)

    def save(self, path):
        with open(path, "wb") as f:
            pickle.dump(self.net.params, f)

    def load(self, path):
        with open(path, "rb") as f:
            params = pickle.load(f)

        self.net.params = params
        for layer in self.net.layers:
            layer.is_init = True

    def get_phase(self):
        return self.net.get_phase()

    def set_phase(self, phase):
        self.net.set_phase(phase)



"""Feed-forward Neural Network class."""

import copy

import numpy as np


class Net(object):

    def __init__(self, layers):
        self.layers = layers
        self._phase = "TRAIN"

    def __repr__(self):
        return "\n".join([str(l) for l in self.layers])

    def forward(self, inputs):
        for layer in self.layers:
            inputs = layer.forward(inputs)
        return inputs

    def backward(self, grad):
        # back propagation
        layer_grads = []
        for layer in reversed(self.layers):
            grad = layer.backward(grad)
            layer_grads.append(copy.copy(layer.grads))

        # return structured gradients
        struct_grad = StructuredParam(layer_grads[::-1])
        # keep the gradient w.r.t the input
        struct_grad.wrt_input = grad
        return struct_grad

    @property
    def params(self):
        return StructuredParam([l.params for l in self.layers])

    @params.setter
    def params(self, params):
        self.params.values = params.values

    def get_phase(self):
        return self._phase

    def set_phase(self, phase):
        for layer in self.layers:
            layer.set_phase(phase)
        self._phase = phase

    def init_params(self, input_shape):
        # manually init params by letting data forward through the network
        self.forward(np.ones((1, *input_shape)))


class StructuredParam(object):
    """A helper class represents network parameters or gradients."""

    def __init__(self, layer_data):
        self.layer_data = layer_data

    @property
    def values(self):
        return np.array([v for d in self.layer_data for v in d.values()])

    @values.setter
    def values(self, values):
        i = 0
        for d in self.layer_data:
            for name in d.keys():
                d[name] = values[i]
                i += 1

    @property
    def shape(self):
        shape = list()
        for d in self.layer_data:
            l_shape = dict()
            for k, v in d.items():
                l_shape[k] = v.shape
            shape.append(l_shape)
        shape = tuple(shape)
        return shape

    @staticmethod
    def _ensure_values(obj):
        if isinstance(obj, StructuredParam):
            obj = obj.values
        return obj

    def clip(self, min_=None, max_=None):
        obj = copy.deepcopy(self)
        obj.values = [v.clip(min_, max_)for v in self.values]
        return obj

    def __add__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values + self._ensure_values(other)
        return obj

    def __radd__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) + self.values
        return obj

    def __iadd__(self, other):
        self.values += self._ensure_values(other)
        return self

    def __sub__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values - self._ensure_values(other)
        return obj

    def __rsub__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) - self.values
        return obj

    def __isub__(self, other):
        other = self._ensure_values(other)
        self.values -= self._ensure_values(other)
        return self

    def __mul__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values * self._ensure_values(other)
        return obj

    def __rmul__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) * self.values
        return obj

    def __imul__(self, other):
        self.values *= self._ensure_values(other)
        return self

    def __truediv__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values / self._ensure_values(other)
        return obj

    def __rtruediv__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) / self.values
        return obj

    def __itruediv__(self, other):
        self.values /= self._ensure_values(other)
        return self

    def __pow__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values ** self._ensure_values(other)
        return obj

    def __ipow__(self, other):
        self.values **= self._ensure_values(other)
        return self

    def __neg__(self):
        obj = copy.deepcopy(self)
        obj.values = -self.values
        return obj

    def __len__(self):
        return len(self.values)

    def __lt__(self, other):
        obj = copy.deepcopy(self)
        other = self._ensure_values(other)

        if isinstance(other, float):
            obj.values = [v < other for v in self.values]
        else:
            obj.values = [v < other[i] for i, v in enumerate(self.values)]
        return obj

    def __gt__(self, other):
        obj = copy.deepcopy(self)
        other = self._ensure_values(other)

        if isinstance(other, float):
            obj.values = [v > other for v in self.values]
        else:
            obj.values = [v > other[i] for i, v in enumerate(self.values)]
        return obj

    def __and__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) & self.values
        return obj

    def __or__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) | self.values
        return obj



"""Common datasets"""

import gzip
import os
import pickle
import sys
import struct
import tarfile

import numpy as np

from tinynn.utils.downloader import download_url


def get_one_hot(targets, nb_classes):
    return np.eye(nb_classes)[np.array(targets).reshape(-1)]


def mnist(data_dir, one_hot=False):
    """
    return: train_set, valid_set, test_set
    train_set size: (50000, 784), (50000,)
    valid_set size: (10000, 784), (10000,)
    test_set size: (10000, 784), (10000,)
    feature: numerical in range [0, 1]
    target: categorical from 0 to 9
    """
    url = "http://deeplearning.net/data/mnist/mnist.pkl.gz"
    checksum = "a02cd19f81d51c426d7ca14024243ce9"

    save_path = os.path.join(data_dir, url.split("/")[-1])
    print("Preparing MNIST dataset ...")
    try:
        download_url(url, save_path, checksum)
    except Exception as e:
        print("Error downloading dataset: %s" % str(e))
        sys.exit(1)

    # load the dataset
    with gzip.open(save_path, "rb") as f:
        train_set, valid_set, test_set = pickle.load(f, encoding="latin1")

    if one_hot:
        train_set = (train_set[0], get_one_hot(train_set[1], 10))
        valid_set = (valid_set[0], get_one_hot(valid_set[1], 10))
        test_set = (test_set[0], get_one_hot(test_set[1], 10))

    return train_set, valid_set, test_set


def fashion_mnist(data_dir, one_hot=False):
    """
    return: train_set, test_set
    train_set size: (60000, 784), (60000,)
    test_set size: (10000, 784), (10000,)
    feature: numerical in range [0, 1]
    target: categorical from 0 to 9
    """
    def read_idx(filename):
        with gzip.open(filename, "rb") as f:
            zero, data_type, dims = struct.unpack(">HBB", f.read(4))
            shape = tuple(struct.unpack(">I", f.read(4))[0] for d in range(dims))
            return np.fromstring(f.read(), dtype=np.uint8).reshape(shape)

    urls = ["http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz",
            "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz",
            "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz",
            "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz"]
    checksums = ["8d4fb7e6c68d591d4c3dfef9ec88bf0d",
                 "25c81989df183df01b3e8a0aad5dffbe",
                 "bef4ecab320f06d8554ea6380940ec79",
                 "bb300cfdad3c16e7a12a480ee83cd310"]
    save_paths = [os.path.join(data_dir, url.split("/")[-1]) for url in urls]

    print("Preparing Fashion-MNIST dataset...")
    for url, checksum, save_path in zip(urls, checksums, save_paths):
        try:
            download_url(url, save_path, checksum)
        except Exception as e:
            print("Error downloading dataset: %s" % str(e))
            sys.exit(1)

    dataset = []
    for path in save_paths:
        dataset.append(read_idx(path))
    train_x, train_y, test_x, test_y = dataset
    # normalize
    train_x = train_x.astype(float) / 255.0
    test_x = test_x.astype(float) / 255.0
    # reshape
    train_x = train_x.reshape((len(train_x), -1))
    test_x = test_x.reshape((len(test_x), -1))
    # one hot if need
    if one_hot:
        train_y = get_one_hot(train_y, 10)
        test_y = get_one_hot(test_y, 10)
    return (train_x, train_y), None, (test_x, test_y)


def cifar10(data_dir, one_hot=False):
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    checksum = "c58f30108f718f92721af3b95e74349a"

    save_path = os.path.join(data_dir, url.split("/")[-1])
    print("Preparing CIFAR-10 dataset...")
    try:
        download_url(url, save_path, checksum)
    except Exception as e:
        print("Error downloading dataset: %s" % str(e))
        sys.exit(1)

    # load the dataset
    dataset = {}
    with open(save_path, "rb") as f:
        tar = tarfile.open(fileobj=f)
        for item in tar:
            obj = tar.extractfile(item)
            if not obj or item.size < 100:
                continue
            cont = pickle.load(obj, encoding="bytes")
            dataset[item.name.split("/")[-1]] = cont

    data_batch_name = ["data_batch_%d" % i for i in range(1, 6)]
    train_x, train_y = [], []
    for name in data_batch_name:
        cont = dataset[name]
        train_x.append(cont[b'data'])
        train_y.extend(cont[b'labels'])
    train_x = np.concatenate(train_x, axis=0)

    # normalize
    means, stds = (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
    train_x = train_x / 255.0
    train_x = train_x.reshape((-1, 1024, 3))
    for c in range(3):
        train_x[:, :, c] = (train_x[:, :, c] - means[c]) / stds[c]
    train_x = train_x.reshape(-1, 3072)

    train_y = np.asarray(train_y)
    train_set = (train_x, train_y)

    test_x = dataset["test_batch"][b"data"]
    test_x = test_x / 255.0
    test_x = test_x.reshape((-1, 1024, 3))
    for c in range(3):
        test_x[:, :, c] = (test_x[:, :, c] - means[c]) / stds[c]
    test_x = test_x.reshape(-1, 3072)
    test_y = np.asarray(dataset["test_batch"][b"labels"])
    test_set = (test_x, test_y)

    if one_hot:
        train_set = (train_set[0], get_one_hot(train_set[1], 10))
        test_set = (test_set[0], get_one_hot(test_set[1], 10))
    return train_set, test_set


def cifar100(data_dir, one_hot=False):
    url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    checksum = "eb9058c3a382ffc7106e4002c42a8d85"

    save_path = os.path.join(data_dir, url.split("/")[-1])
    print("Preparing CIFAR-100 dataset...")
    try:
        download_url(url, save_path, checksum)
    except Exception as e:
        print("Error downloading dataset: %s" % str(e))
        sys.exit(1)

    # load the dataset
    dataset = {}
    with open(save_path, "rb") as f:
        tar = tarfile.open(fileobj=f)
        for item in tar:
            obj = tar.extractfile(item)
            if not obj or item.size < 100:
                continue
            cont = pickle.load(obj, encoding="bytes")
            dataset[item.name.split("/")[-1]] = cont

    train_x = dataset["train"][b"data"]
    train_x = train_x / 255.0
    train_y = np.asarray(dataset["train"][b"fine_labels"])
    train_set = (train_x, train_y)

    test_x = dataset["test"][b"data"]
    test_x = test_x / 255.0
    test_y = np.asarray(dataset["test"][b"fine_labels"])
    test_set = (test_x, test_y)

    if one_hot:
        train_set = (train_set[0], get_one_hot(train_set[1], 10))
        test_set = (test_set[0], get_one_hot(test_set[1], 10))
    return train_set, test_set



"""Network layers."""

import numpy as np
from tinynn.core.initializer import Ones
from tinynn.core.initializer import XavierUniform
from tinynn.core.initializer import Zeros


class Layer(object):

    def __init__(self):
        self.params = {p: None for p in self.param_names}
        self.grads = {}
        self.shapes = {}

        self.is_training = True
        self.is_init = False

    def forward(self, inputs):
        raise NotImplementedError

    def backward(self, grad):
        raise NotImplementedError

    def set_phase(self, phase):
        self.is_training = True if phase == "TRAIN" else False

    @property
    def name(self):
        return self.__class__.__name__

    def __repr__(self):
        return "layer: %s \t shape: %s" % (self.name, self.shapes)

    @property
    def param_names(self):
        return ()


class Dense(Layer):

    def __init__(self,
                 num_out,
                 w_init=XavierUniform(),
                 b_init=Zeros()):
        super().__init__()

        self.initializers = {"w": w_init, "b": b_init}
        self.shapes = {"w": [None, num_out], "b": [num_out]}

        self.inputs = None

    def forward(self, inputs):
        if not self.is_init:
            self.shapes["w"][0] = inputs.shape[1]
            self._init_params()
        self.inputs = inputs
        return inputs @ self.params["w"] + self.params["b"]

    def backward(self, grad):
        self.grads["w"] = self.inputs.T @ grad
        self.grads["b"] = np.sum(grad, axis=0)
        return grad @ self.params["w"].T

    def _init_params(self):
        for p in self.param_names:
            self.params[p] = self.initializers[p](self.shapes[p])
        self.is_init = True

    @property
    def param_names(self):
        return "w", "b"


class Conv2D(Layer):
    """
    Implement 2D convolution layer
    :param kernel: A list/tuple of int that has length 4 (height, width,
        in_channels, out_channels)
    :param stride: A list/tuple of int that has length 2 (height, width)
    :param padding: String ["SAME", "VALID"]
    :param w_init: weight initializer
    :param b_init: bias initializer
    """
    def __init__(self,
                 kernel,
                 stride=(1, 1),
                 padding="SAME",
                 w_init=XavierUniform(),
                 b_init=Zeros()):
        super().__init__()

        self.kernel_shape = kernel
        self.stride = stride
        self.initializers = {"w": w_init, "b": b_init}
        self.shapes = {"w": self.kernel_shape, "b": self.kernel_shape[-1]}

        self.padding_mode = padding
        self.padding = None

    def forward(self, inputs):
        """
        Accelerate convolution via im2col trick.
        An example (assuming only one channel and one filter):
         input = | 43  16  78 |         kernel = | 4  6 |
          (X)    | 34  76  95 |                  | 7  9 |
                 | 35   8  46 |
        
        After im2col and kernel flattening:
         col  = | 43  16  34  76 |     kernel = | 4 |
                | 16  78  76  95 |      (W)     | 6 |
                | 34  76  35   8 |              | 7 |
                | 76  95   8  46 |              | 9 |
        """
        if not self.is_init:
            self._init_params()

        k_h, k_w, in_c, out_c = self.kernel_shape
        s_h, s_w = self.stride
        X = self._inputs_preprocess(inputs)

        # padded inputs to column matrix
        col = im2col(X, k_h, k_w, s_h, s_w)
        # perform convolution by matrix product.
        W = self.params["w"].reshape(-1, out_c)
        Z = col @ W
        # reshape output 
        batch_sz, in_h, in_w, _ = X.shape
        # separate the batch size and feature map dimensions
        Z = Z.reshape(batch_sz, Z.shape[0] // batch_sz, out_c)
        # further divide the feature map in to (h, w) dimension
        out_h = (in_h - k_h) // s_h + 1
        out_w = (in_w - k_w) // s_w + 1
        Z = Z.reshape(batch_sz, out_h, out_w, out_c)

        # plus the bias for every filter
        Z += self.params["b"]
        # save results for backward function
        self.X_shape, self.col, self.W = X.shape, col, W
        return Z

    def backward(self, grad):
        """
        Compute gradients w.r.t layer parameters and backward gradients.
        :param grad: gradients from previous layer 
            with shape (batch_sz, out_h, out_w, out_c)
        :return d_in: gradients to next layers 
            with shape (batch_sz, in_h, in_w, in_c)
        """
        # read size parameters
        k_h, k_w, in_c, out_c = self.kernel_shape
        s_h, s_w = self.stride
        batch_sz, in_h, in_w, in_c = self.X_shape
        pad_h, pad_w = self.padding[1:3]

        # grads w.r.t parameters
        flat_grad = grad.reshape((-1, out_c))
        d_W = self.col.T @ flat_grad
        self.grads["w"] = d_W.reshape(self.kernel_shape)
        self.grads["b"] = np.sum(flat_grad, axis=0)

        # grads w.r.t inputs
        d_X = grad @ self.W.T
        # cast gradients back to original shape as d_in
        d_in = np.zeros(shape=self.X_shape)
        for i, r in enumerate(range(0, in_h - k_h + 1, s_h)):
            for j, c in enumerate(range(0, in_w - k_w + 1, s_w)):
                patch = d_X[:, i, j, :]
                patch = patch.reshape((batch_sz, k_h, k_w, in_c))
                d_in[:, r:r+k_h, c:c+k_w, :] += patch
                
        # cut off gradients of padding
        d_in = d_in[:, pad_h[0]:in_h-pad_h[1], pad_w[0]:in_w-pad_w[1], :]
        return self._grads_postprocess(d_in)

    def _inputs_preprocess(self, inputs):
        _, in_h, in_w, _ = inputs.shape
        k_h, k_w, _, _ = self.kernel_shape
        # padding calculation
        if self.padding is None:
            self.padding = get_padding_2d(
                (in_h, in_w), (k_h, k_w), self.padding_mode)
        return np.pad(inputs, pad_width=self.padding, mode="constant")

    def _grads_postprocess(self, grads):
        return grads

    def _init_params(self):
        for p in self.param_names:
            self.params[p] = self.initializers[p](self.shapes[p])
        self.is_init = True

    @property
    def param_names(self):
        return "w", "b"


class ConvTranspose2D(Conv2D):

    def __init__(self,
                 kernel,
                 stride=(1, 1),
                 padding="SAME",
                 w_init=XavierUniform(),
                 b_init=Zeros()):
        super().__init__(kernel, stride, padding, w_init, b_init)
        self.origin_stride = stride
        self.stride = (1, 1)

    def _inputs_preprocess(self, inputs):
        k_h, k_w = self.kernel_shape[:2]
        # insert zeros to inputs
        inputs = self._insert_zeros(
            inputs, *self.origin_stride, self.padding_mode)
        batch_sz, in_h, in_w, in_c = inputs.shape
        # padding calculation
        if self.padding is None:
            if self.padding_mode == "SAME":
                self.padding = get_padding_2d(
                    (in_h, in_w), (k_h, k_w), self.padding_mode)
            else:
                self.padding = ((0, 0), (k_h - 1, k_h - 1),
                                (k_w - 1, k_w - 1), (0, 0))
        return np.pad(inputs, pad_width=self.padding, mode="constant")

    def _grads_postprocess(self, grads):
        return grads[:, ::self.origin_stride[0], ::self.origin_stride[1], :]

    @staticmethod
    def _insert_zeros(inputs, s_h, s_w, mode):
        batch_sz, in_h, in_w, in_c = inputs.shape
        if mode == "SAME":
            out_h = in_h * s_h
            out_w = in_w * s_w
        else:
            out_h = (in_h - 1) * s_h + 1
            out_w = (in_w - 1) * s_h + 1
        expand = np.zeros((batch_sz, out_h, out_w, in_c))
        expand[:, ::s_h, ::s_w, :] = inputs
        return expand


class MaxPool2D(Layer):

    def __init__(self, pool_size, stride, padding="VALID"):
        """
        Implement 2D max-pooling layer
        :param pool_size: A list/tuple of 2 integers (pool_height, pool_width)
        :param stride: A list/tuple of 2 integers (stride_height, stride_width)
        :param padding: A string ("SAME", "VALID")
        """
        super().__init__()
        self.kernel_shape = pool_size
        self.stride = stride

        self.padding_mode = padding
        self.padding = None

    def forward(self, inputs):
        s_h, s_w = self.stride
        k_h, k_w = self.kernel_shape
        batch_sz, in_h, in_w, in_c = inputs.shape

        # zero-padding
        if self.padding is None:
            self.padding = get_padding_2d(
                (in_h, in_w), (k_h, k_w), self.padding_mode)
        X = np.pad(inputs, pad_width=self.padding, mode="constant")
        padded_h, padded_w = X.shape[1:3]
    
        out_h = (padded_h - k_h) // s_h + 1
        out_w = (padded_w - k_w) // s_w + 1

        # construct output matrix and argmax matrix
        max_pool = np.empty(shape=(batch_sz, out_h, out_w, in_c))
        argmax = np.empty(shape=(batch_sz, out_h, out_w, in_c), dtype=int)
        for r in range(out_h):
            r_start = r * s_h
            for c in range(out_w):
                c_start = c * s_w
                pool = X[:, r_start: r_start+k_h, c_start: c_start+k_w, :]
                pool = pool.reshape((batch_sz, -1, in_c))

                _argmax = np.argmax(pool, axis=1)[:, np.newaxis, :]
                argmax[:, r, c, :] = _argmax.squeeze()

                # get max elements
                _max_pool = np.take_along_axis(pool, _argmax, axis=1).squeeze()
                max_pool[:, r, c, :] = _max_pool

        self.X_shape = X.shape
        self.out_shape = (out_h, out_w)
        self.argmax = argmax
        return max_pool

    def backward(self, grad):
        batch_sz, in_h, in_w, in_c = self.X_shape
        out_h, out_w = self.out_shape
        s_h, s_w = self.stride
        k_h, k_w = self.kernel_shape
        k_sz = k_h * k_w
        pad_h, pad_w = self.padding[1:3]

        d_in = np.zeros(shape=(batch_sz, in_h, in_w, in_c))
        for r in range(out_h):
            r_start = r * s_h
            for c in range(out_w):
                c_start = c * s_w
                _argmax = self.argmax[:, r, c, :]
                mask = np.eye(k_sz)[_argmax].transpose((0, 2, 1))
                _grad = grad[:, r, c, :][:, np.newaxis, :]
                patch = np.repeat(_grad, k_sz, axis=1) * mask
                patch = patch.reshape((batch_sz, k_h, k_w, in_c))
                d_in[:, r_start:r_start+k_h, c_start:c_start+k_w, :] += patch

        # cut off gradients of padding
        d_in = d_in[:, pad_h[0]:in_h-pad_h[1], pad_w[0]:in_w-pad_w[1], :]
        return d_in


class RNN(Layer):
    
    def __init__(self, 
                 num_hidden,
                 activation,
                 bptt_trunc=None,
                 w_init=XavierUniform(),
                 b_init=Zeros()):
        super().__init__()
        self.num_hidden = num_hidden
        self.activation = activation
        self.bptt_trunc = bptt_trunc

        self.initializer = {"W": w_init, "V": w_init, "U": w_init,
                            "b": b_init, "c": b_init}

    def forward(self, inputs):
        """
        Vanilla recurrent neural net forward pass
        a_{t} = U @ x_{t} + W @ s_{t-1} + b
        h_{t} = activation_func(a_{t}) 
        o_{t} = V @ h_{t} + c
        """
        batch_size, n_ts, input_dim = inputs.shape
        if not self.is_init:
            self.shapes = {"W": [self.num_hidden, self.num_hidden],
                           "V": [input_dim, self.num_hidden],
                           "U": [self.num_hidden, input_dim],
                           "b": [self.num_hidden],
                           "c": [input_dim]}
            self._init_params()

        a = np.empty((batch_size, n_ts, self.num_hidden))
        h = np.empty((batch_size, n_ts + 1, self.num_hidden))
        out = np.empty((batch_size, n_ts, input_dim))

        h[:, -1] = np.zeros((batch_size, self.num_hidden))
        for t in range(n_ts):
            a[:, t] = (inputs[:, t] @ self.params["U"].T +
                       h[:, t-1] @ self.params["W"].T + self.params["b"])
            h[:, t] = self.activation.forward(a[:, t])
            out[:, t] = h[:, t] @ self.params["V"].T + self.params["c"]

        # cache for backward pass
        self.h, self.a, self.X = h, a, inputs
        return out[:, -1]

    def backward(self, grad):
        n_ts = self.X.shape[1]
        for p in self.param_names:
            self.grads[p] = np.zeros_like(self.params[p])

        if self.bptt_trunc is None:
            self.bptt_trunc = n_ts  # non-truncated
        
        d_in = np.empty_like(self.X)
        for t in reversed(range(n_ts)):
            # grads w.r.t param V and c
            self.grads["c"] += grad.sum(axis=0)
            self.grads["V"] += grad.T @ self.h[:, t]
            # grads w.r.t h
            d_h = grad @ self.params["V"]
            d_a = d_h * self.activation.derivative(self.a[:, t])
            # grads w.r.t input X
            d_in[:, t] = d_a @ self.params["U"]
            # grads w.r.t params U, W and b
            for i in range(min(self.bptt_trunc, t+1)):
                self.grads["U"] += d_a.T @ self.X[:, t - i]
                self.grads["W"] += d_a.T @ self.h[:, t - i - 1]
                self.grads["b"] += d_a.sum(axis=0)
                d_h = d_a @ self.params["W"]
                d_a = d_h * self.activation.derivative(self.a[:, t - i - 1])
        return d_in

    def _init_params(self):
        for p in self.param_names:
            self.params[p] = self.initializer[p](self.shapes[p])
        self.is_init = True

    @property
    def param_names(self):
        return "W", "U", "V", "b", "c"


class BatchNormalization(Layer):

    def __init__(self,
                 momentum=0.99,
                 gamma_init=Ones(),
                 beta_init=Zeros(),
                 running_mean_init=Zeros(),
                 running_var_init=Ones(),
                 epsilon=1e-5):
        super().__init__()
        self.m = momentum
        self.epsilon = epsilon

        self.initializer = {"gamma": gamma_init, 
                            "beta": beta_init,
                            "r_mean": running_mean_init,
                            "r_var": running_var_init}

    def forward(self, inputs):
        if not self.is_init:
            for p in self.param_names:
                self.shapes[p] = inputs.shape[1:]
            self._init_params()

        if self.is_training:
            mean = inputs.mean(axis=0)
            var = inputs.var(axis=0)
            self.params["r_mean"] = (self.m * self.params["r_mean"] +
                                     (1.0 - self.m) * mean)
            self.params["r_var"] = (self.m * self.params["r_var"] +
                                    (1.0 - self.m) * var)
        else:
            n = inputs.shape[0]
            mean = self.params["r_mean"]
            # unbiased variance estimate
            var = (n / (n - 1)) * self.params["r_var"]

        # normalize
        self.X_center = inputs - mean
        self.std = (var + self.epsilon) ** 0.5
        self.X_norm = self.X_center / self.std
        return self.params["gamma"] * self.X_norm + self.params["beta"]

    def backward(self, grad):
        # BUG: The optimizer will change r_mean and r_var
        #  even though their gradients are zeros.
        self.grads["r_mean"] = 0
        self.grads["r_var"] = 0

        self.grads["gamma"] = (self.X_norm * grad).sum(axis=0)
        self.grads["beta"] = grad.sum(axis=0)

        batch_size = grad.shape[0]
        std_inv = 1.0 / self.std
        # grads w.r.t inputs
        # ref: http://cthorey.github.io./backpropagation/
        d_in = (1.0 / batch_size) * self.params["gamma"] * std_inv * (
            batch_size * grad - np.sum(grad, axis=0) - 
            self.X_center * std_inv ** 2 * np.sum(grad * self.X_center, axis=0))
        return d_in

    def _init_params(self):
        for p in self.param_names:
            self.params[p] = self.initializer[p](self.shapes[p])
        self.is_init = True

    @property
    def param_names(self):
        return "gamma", "beta", "r_mean", "r_var"


class Reshape(Layer):

    def __init__(self, *output_shape):
        super().__init__()
        self.output_shape = output_shape
        self.input_shape = None

    def forward(self, inputs):
        self.input_shape = inputs.shape
        return inputs.reshape(inputs.shape[0], *self.output_shape)

    def backward(self, grad):
        return grad.reshape(self.input_shape)


class Flatten(Reshape):

    def __init__(self):
        super().__init__(-1)


class Dropout(Layer):

    def __init__(self, keep_prob=0.5):
        super().__init__()
        self._keep_prob = keep_prob
        self._multiplier = None

    def forward(self, inputs):
        if self.is_training:
            multiplier = np.random.binomial(
                1, self._keep_prob, size=inputs.shape)
            self._multiplier = multiplier / self._keep_prob
            outputs = inputs * self._multiplier
        else:
            outputs = inputs
        return outputs

    def backward(self, grad):
        assert self.is_training is True
        return grad * self._multiplier


class Activation(Layer):

    def __init__(self):
        super().__init__()
        self.inputs = None

    def forward(self, inputs):
        self.inputs = inputs
        return self.func(inputs)

    def backward(self, grad):
        return self.derivative(self.inputs) * grad

    def func(self, x):
        raise NotImplementedError

    def derivative(self, x):
        raise NotImplementedError


class Sigmoid(Activation):

    def func(self, x):
        return 1.0 / (1.0 + np.exp(-x))

    def derivative(self, x):
        return self.func(x) * (1.0 - self.func(x))


class Softplus(Activation):

    def func(self, x):
        return np.log(1.0 + np.exp(-np.abs(x))) + np.maximum(x, 0.0)

    def derivative(self, x):
        return 1.0 / (1.0 + np.exp(-x))


class Tanh(Activation):

    def func(self, x):
        return np.tanh(x)

    def derivative(self, x):
        return 1.0 - self.func(x) ** 2


class ReLU(Activation):

    def func(self, x):
        return np.maximum(x, 0.0)

    def derivative(self, x):
        return x > 0.0


class LeakyReLU(Activation):

    def __init__(self, slope=0.2):
        super().__init__()
        self._slope = slope

    def func(self, x):
        x = x.copy()
        x[x < 0.0] *= self._slope
        return x

    def derivative(self, x):
        dx = np.ones_like(x)
        dx[x < 0.0] = self._slope
        return dx


def im2col(img, k_h, k_w, s_h, s_w):
    """
    Transform padded image into column matrix.
    :param img: padded inputs of shape (B, in_h, in_w, in_c)
    :param k_h: kernel height
    :param k_w: kernel width
    :param s_h: stride height
    :param s_w: stride width
    :return col: column matrix of shape (B*out_h*out_w, k_h*k_h*inc)
    """
    batch_sz, h, w, in_c = img.shape
    # calculate result feature map size
    out_h = (h - k_h) // s_h + 1
    out_w = (w - k_w) // s_w + 1
    # allocate space for column matrix
    col = np.empty((batch_sz * out_h * out_w, k_h * k_w * in_c))
    # fill in the column matrix
    batch_span = out_w * out_h
    for r in range(out_h):
        r_start = r * s_h
        matrix_r = r * out_w
        for c in range(out_w):
            c_start = c * s_w
            patch = img[:, r_start: r_start+k_h, c_start: c_start+k_w, :]
            patch = patch.reshape(batch_sz, -1)
            col[matrix_r+c::batch_span, :] = patch
    return col


def get_padding_2d(in_shape, k_shape, mode):
    def get_padding_1d(w, k):
        if mode == "SAME":
            pads = (w - 1) + k - w
            half = pads // 2
            padding = (half, half) if pads % 2 == 0 else (half, half + 1)
        else:
            padding = (0, 0)
        return padding

    h_pad = get_padding_1d(in_shape[0], k_shape[0])
    w_pad = get_padding_1d(in_shape[1], k_shape[1])
    return (0, 0), h_pad, w_pad, (0, 0)



"""Data Iterator class."""

from collections import namedtuple

import numpy as np

Batch = namedtuple("Batch", ["inputs", "targets"])


class BaseIterator(object):

    def __call__(self, inputs, targets):
        raise NotImplementedError


class BatchIterator(BaseIterator):

    def __init__(self, batch_size=32, shuffle=True):
        self.batch_size = batch_size
        self.shuffle = shuffle

    def __call__(self, inputs, targets):
        starts = np.arange(0, len(inputs), self.batch_size)
        if self.shuffle:
            idx = np.arange(len(inputs))
            np.random.shuffle(idx)
            inputs = inputs[idx]
            targets = targets[idx]

        for start in starts:
            end = start + self.batch_size
            batch_inputs = inputs[start: end]
            batch_targets = targets[start: end]
            yield Batch(inputs=batch_inputs, targets=batch_targets)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tinynn",
    version="0.0.4",
    author="borgwang",
    author_email="borgwang@126.com",
    description="A lightweight deep learning library",
    long_description="tinynn is an open source, lightweight deep learning library "
                     "written in Python.",
    long_description_content_type="text/markdown",
    url="https://github.com/borgwang/tinynn",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    install_requires=[
        "numpy>=1.16.0",
        "scipy>=1.3.1"
    ],
    python_requires='>=3.6'
)



"""Loss functions"""

import numpy as np
from tinynn.utils.math import log_softmax


class Loss(object):

    def loss(self, *args, **kwargs):
        raise NotImplementedError

    def grad(self, *args, **kwargs):
        raise NotImplementedError


class MSE(Loss):

    def loss(self, predicted, actual):
        m = predicted.shape[0]
        return 0.5 * np.sum((predicted - actual) ** 2) / m

    def grad(self, predicted, actual):
        m = predicted.shape[0]
        return (predicted - actual) / m


class MAE(Loss):

    def loss(self, predicted, actual):
        m = predicted.shape[0]
        return np.sum(np.abs(predicted - actual)) / m

    def grad(self, predicted, actual):
        m = predicted.shape[0]
        return np.sign(predicted - actual) / m


class Huber(Loss):

    def __init__(self, delta=1.0):
        self._delta = delta

    def loss(self, predicted, actual):
        l1_dist = np.abs(predicted - actual)
        mse_mask = l1_dist < self._delta  # MSE part
        mae_mask = ~mse_mask  # MAE part
        mse = 0.5 * (predicted - actual) ** 2
        mae = self._delta * np.abs(predicted - actual) - 0.5 * self._delta ** 2

        m = predicted.shape[0]
        return np.sum(mse * mse_mask + mae * mae_mask) / m

    def grad(self, predicted, actual):
        err = predicted - actual
        mse_mask = np.abs(err) < self._delta  # MSE part
        mae_mask = ~mse_mask  # MAE part
        m = predicted.shape[0]
        mse_grad = err / m
        mae_grad = np.sign(err) / m
        return (mae_grad * mae_mask + mse_grad * mse_mask) / m


class SoftmaxCrossEntropy(Loss):

    def __init__(self, T=1.0, weight=None):
        """
        L = weight[class] * (-log(exp(x[class]) / sum(exp(x))))
        :paras T: temperature
        :param weight: A 1D tensor [n_classes] assigning weight to each corresponding sample.
        """
        weight = np.asarray(weight) if weight is not None else weight
        self._weight = weight
        self._T = T

    def loss(self, logits, labels):
        m = logits.shape[0]
        nll = -(log_softmax(logits, t=self._T, axis=1) * labels).sum(axis=1)

        if self._weight is not None:
            nll *= self._weight[labels]
        return np.sum(nll) / m

    def grad(self, logits, labels):
        m = logits.shape[0]
        grad = np.copy(logits)
        grad -= labels
        return grad / m


class SigmoidCrossEntropy(Loss):
    """
    logits = a, label = y
    L = -y * log(1 / (1 + exp(-a)) - (1-y) * log(exp(-a) / (1 + exp(-a))
      = -y * a + log(1 + exp(a))

    In order to get stable version, we can further derive
    L = -y * a + log((1 + exp(-a)) / exp(-a))
      = -y * a + log(1 + exp(-a)) + a
    """
    def __init__(self, weight=None):
        weight = np.asarray(weight) if weight is not None else weight
        self._weight = weight

    def loss(self, logits, labels):
        m = logits.shape[0]
        cost = -labels * logits + np.log(1 + np.exp(-logits)) + logits
        return np.sum(cost) / m

    def grad(self, logits, labels):
        m = logits.shape[0]
        grad = -labels + 1.0 / (1 + np.exp(-logits))
        return grad / m
